@@ -21,17 +21,6 @@ import (
 	"testing"
 )
 
-// testcovdata returns the path to the unit test executable to be used as
-// standin for 'go tool covdata'.
-func testcovdata(t testing.TB) string {
-	exe, err := os.Executable()
-	if err != nil {
-		t.Helper()
-		t.Fatal(err)
-	}
-	return exe
-}
-
 // Top level tempdir for test.
 var testTempDir string
 
@@ -184,7 +173,7 @@ func TestCovTool(t *testing.T) {
 	s.exepath3, s.exedir3 = buildProg(t, "prog1", dir, "atomic", flags)
 
 	// Reuse unit test executable as tool to be tested.
-	s.tool = testcovdata(t)
+	s.tool = testenv.Executable(t)
 
 	// Create a few coverage output dirs.
 	for i := 0; i < 4; i++ {
@@ -294,7 +283,7 @@ func runToolOp(t *testing.T, s state, op string, args []string) []string {
 	cmd := testenv.Command(t, s.tool, args...)
 	b, err := cmd.CombinedOutput()
 	if err != nil {
-		fmt.Fprintf(os.Stderr, "## %s output: %s\n", op, string(b))
+		fmt.Fprintf(os.Stderr, "## %s output: %s\n", op, b)
 		t.Fatalf("%q run error: %v", op, err)
 	}
 	output := strings.TrimSpace(string(b))
