@@ -10,8 +10,9 @@ import (
 	"cmd/go/internal/cache"
 	"cmd/go/internal/cfg"
 	"cmd/go/internal/load"
-	"cmd/go/internal/par"
 	"cmd/go/internal/str"
+	"cmd/internal/par"
+	"cmd/internal/pathcache"
 	"errors"
 	"fmt"
 	"internal/lazyregexp"
@@ -42,7 +43,7 @@ type shellShared struct {
 	workDir string // $WORK, immutable
 
 	printLock sync.Mutex
-	printFunc func(args ...any) (int, error)
+	printer   load.Printer
 	scriptDir string // current directory in printed script
 
 	mkdirCache par.Cache[string, error] // a cache of created directories
@@ -50,31 +51,43 @@ type shellShared struct {
 
 // NewShell returns a new Shell.
 //
-// Shell will internally serialize calls to the print function.
-// If print is nil, it defaults to printing to stderr.
-func NewShell(workDir string, print func(a ...any) (int, error)) *Shell {
-	if print == nil {
-		print = func(a ...any) (int, error) {
-			return fmt.Fprint(os.Stderr, a...)
-		}
+// Shell will internally serialize calls to the printer.
+// If printer is nil, it uses load.DefaultPrinter.
+func NewShell(workDir string, printer load.Printer) *Shell {
+	if printer == nil {
+		printer = load.DefaultPrinter()
 	}
 	shared := &shellShared{
-		workDir:   workDir,
-		printFunc: print,
+		workDir: workDir,
+		printer: printer,
 	}
 	return &Shell{shellShared: shared}
 }
 
-// Print emits a to this Shell's output stream, formatting it like fmt.Print.
-// It is safe to call concurrently.
-func (sh *Shell) Print(a ...any) {
-	sh.printLock.Lock()
-	defer sh.printLock.Unlock()
-	sh.printFunc(a...)
+func (sh *Shell) pkg() *load.Package {
+	if sh.action == nil {
+		return nil
+	}
+	return sh.action.Package
 }
 
-func (sh *Shell) printLocked(a ...any) {
-	sh.printFunc(a...)
+// Printf emits a to this Shell's output stream, formatting it like fmt.Printf.
+// It is safe to call concurrently.
+func (sh *Shell) Printf(format string, a ...any) {
+	sh.printLock.Lock()
+	defer sh.printLock.Unlock()
+	sh.printer.Printf(sh.pkg(), format, a...)
+}
+
+func (sh *Shell) printfLocked(format string, a ...any) {
+	sh.printer.Printf(sh.pkg(), format, a...)
+}
+
+// Errorf reports an error on sh's package and sets the process exit status to 1.
+func (sh *Shell) Errorf(format string, a ...any) {
+	sh.printLock.Lock()
+	defer sh.printLock.Unlock()
+	sh.printer.Errorf(sh.pkg(), format, a...)
 }
 
 // WithAction returns a Shell identical to sh, but bound to Action a.
@@ -358,7 +371,7 @@ func (sh *Shell) ShowCmd(dir string, format string, args ...any) {
 	if dir != "" && dir != "/" {
 		if dir != sh.scriptDir {
 			// Show changing to dir and update the current directory.
-			sh.printLocked(sh.fmtCmd("", "cd %s\n", dir))
+			sh.printfLocked("%s", sh.fmtCmd("", "cd %s\n", dir))
 			sh.scriptDir = dir
 		}
 		// Replace scriptDir is our working directory. Replace it
@@ -370,7 +383,7 @@ func (sh *Shell) ShowCmd(dir string, format string, args ...any) {
 		cmd = strings.ReplaceAll(" "+cmd, " "+dir, dot)[1:]
 	}
 
-	sh.printLocked(cmd + "\n")
+	sh.printfLocked("%s\n", cmd)
 }
 
 // reportCmd reports the output and exit status of a command. The cmdOut and
@@ -509,7 +522,7 @@ func (sh *Shell) reportCmd(desc, dir string, cmdOut []byte, cmdErr error) error
 		a.output = append(a.output, err.Error()...)
 	} else {
 		// Write directly to the Builder output.
-		sh.Print(err.Error())
+		sh.Printf("%s", err)
 	}
 	return nil
 }
@@ -606,7 +619,7 @@ func (sh *Shell) runOut(dir string, env []string, cmdargs ...any) ([]byte, error
 	}
 
 	var buf bytes.Buffer
-	path, err := cfg.LookPath(cmdline[0])
+	path, err := pathcache.LookPath(cmdline[0])
 	if err != nil {
 		return nil, err
 	}
