@@ -35,9 +35,85 @@ import (
 	"cmd/internal/src"
 	"cmd/internal/sys"
 	"internal/abi"
+	"internal/buildcfg"
 	"log"
+	"math"
+	"math/bits"
+	"strings"
 )
 
+// Test if this value can encoded as a mask for
+// li -1, rx; rlic rx,rx,sh,mb.
+// Masks can also extend from the msb and wrap to
+// the lsb too. That is, the valid masks are 32 bit strings
+// of the form: 0..01..10..0 or 1..10..01..1 or 1...1
+func isPPC64DoublewordRotateMask(v64 int64) bool {
+	// Isolate rightmost 1 (if none 0) and add.
+	v := uint64(v64)
+	vp := (v & -v) + v
+	// Likewise, for the wrapping case.
+	vn := ^v
+	vpn := (vn & -vn) + vn
+	return (v&vp == 0 || vn&vpn == 0) && v != 0
+}
+
+// Encode a doubleword rotate mask into mb (mask begin) and
+// me (mask end, inclusive). Note, POWER ISA labels bits in
+// big endian order.
+func encodePPC64RLDCMask(mask int64) (mb, me int) {
+	// Determine boundaries and then decode them
+	mb = bits.LeadingZeros64(uint64(mask))
+	me = 64 - bits.TrailingZeros64(uint64(mask))
+	mbn := bits.LeadingZeros64(^uint64(mask))
+	men := 64 - bits.TrailingZeros64(^uint64(mask))
+	// Check for a wrapping mask (e.g bits at 0 and 63)
+	if mb == 0 && me == 64 {
+		// swap the inverted values
+		mb, me = men, mbn
+	}
+	// Note, me is inclusive.
+	return mb, me - 1
+}
+
+// Is this a symbol which should never have a TOC prologue generated?
+// These are special functions which should not have a TOC regeneration
+// prologue.
+func isNOTOCfunc(name string) bool {
+	switch {
+	case name == "runtime.duffzero":
+		return true
+	case name == "runtime.duffcopy":
+		return true
+	case strings.HasPrefix(name, "runtime.elf_"):
+		return true
+	default:
+		return false
+	}
+}
+
+// Try converting FMOVD/FMOVS to XXSPLTIDP. If it is converted,
+// return true.
+func convertFMOVtoXXSPLTIDP(p *obj.Prog) bool {
+	if p.From.Type != obj.TYPE_FCONST || buildcfg.GOPPC64 < 10 {
+		return false
+	}
+	v := p.From.Val.(float64)
+	if float64(float32(v)) != v {
+		return false
+	}
+	// Secondly, is this value a normal value?
+	ival := int64(math.Float32bits(float32(v)))
+	isDenorm := ival&0x7F800000 == 0 && ival&0x007FFFFF != 0
+	if !isDenorm {
+		p.As = AXXSPLTIDP
+		p.From.Type = obj.TYPE_CONST
+		p.From.Offset = ival
+		// Convert REG_Fx into equivalent REG_VSx
+		p.To.Reg = REG_VS0 + (p.To.Reg & 31)
+	}
+	return !isDenorm
+}
+
 func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 	p.From.Class = 0
 	p.To.Class = 0
@@ -59,7 +135,7 @@ func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 	// Rewrite float constants to values stored in memory.
 	switch p.As {
 	case AFMOVS:
-		if p.From.Type == obj.TYPE_FCONST {
+		if p.From.Type == obj.TYPE_FCONST && !convertFMOVtoXXSPLTIDP(p) {
 			f32 := float32(p.From.Val.(float64))
 			p.From.Type = obj.TYPE_MEM
 			p.From.Sym = ctxt.Float32Sym(f32)
@@ -71,7 +147,7 @@ func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 		if p.From.Type == obj.TYPE_FCONST {
 			f64 := p.From.Val.(float64)
 			// Constant not needed in memory for float +/- 0
-			if f64 != 0 {
+			if f64 != 0 && !convertFMOVtoXXSPLTIDP(p) {
 				p.From.Type = obj.TYPE_MEM
 				p.From.Sym = ctxt.Float64Sym(f64)
 				p.From.Name = obj.NAME_EXTERN
@@ -79,16 +155,78 @@ func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 			}
 		}
 
+	case AMOVW, AMOVWZ:
+		// Note, for backwards compatibility, MOVW $const, Rx and MOVWZ $const, Rx are identical.
+		if p.From.Type == obj.TYPE_CONST && p.From.Offset != 0 && p.From.Offset&0xFFFF == 0 {
+			// This is a constant shifted 16 bits to the left, convert it to ADDIS/ORIS $const,...
+			p.As = AADDIS
+			// Use ORIS for large constants which should not be sign extended.
+			if p.From.Offset >= 0x80000000 {
+				p.As = AORIS
+			}
+			p.Reg = REG_R0
+			p.From.Offset >>= 16
+		}
+
 	case AMOVD:
-		// 32b constants (signed and unsigned) can be generated via 1 or 2 instructions.
-		// All others must be placed in memory and loaded.
+		// Skip this opcode if it is not a constant load.
+		if p.From.Type != obj.TYPE_CONST || p.From.Name != obj.NAME_NONE || p.From.Reg != 0 {
+			break
+		}
+
+		// 32b constants (signed and unsigned) can be generated via 1 or 2 instructions. They can be assembled directly.
 		isS32 := int64(int32(p.From.Offset)) == p.From.Offset
 		isU32 := uint64(uint32(p.From.Offset)) == uint64(p.From.Offset)
-		if p.From.Type == obj.TYPE_CONST && p.From.Name == obj.NAME_NONE && p.From.Reg == 0 && !isS32 && !isU32 {
-			p.From.Type = obj.TYPE_MEM
-			p.From.Sym = ctxt.Int64Sym(p.From.Offset)
-			p.From.Name = obj.NAME_EXTERN
-			p.From.Offset = 0
+		// If prefixed instructions are supported, a 34b signed constant can be generated by one pli instruction.
+		isS34 := pfxEnabled && (p.From.Offset<<30)>>30 == p.From.Offset
+
+		// Try converting MOVD $const,Rx into ADDIS/ORIS $s32>>16,R0,Rx
+		switch {
+		case isS32 && p.From.Offset&0xFFFF == 0 && p.From.Offset != 0:
+			p.As = AADDIS
+			p.From.Offset >>= 16
+			p.Reg = REG_R0
+
+		case isU32 && p.From.Offset&0xFFFF == 0 && p.From.Offset != 0:
+			p.As = AORIS
+			p.From.Offset >>= 16
+			p.Reg = REG_R0
+
+		case isS32 || isU32 || isS34:
+			// The assembler can generate this opcode in 1 (on Power10) or 2 opcodes.
+
+		// Otherwise, see if the large constant can be generated with 2 instructions. If not, load it from memory.
+		default:
+			// Is this a shifted 16b constant? If so, rewrite it to avoid a creating and loading a constant.
+			val := p.From.Offset
+			shift := bits.TrailingZeros64(uint64(val))
+			mask := int64(0xFFFF) << shift
+			if val&mask == val || (val>>(shift+16) == -1 && (val>>shift)<<shift == val) {
+				// Rewrite this value into MOVD $const>>shift, Rto; SLD $shift, Rto
+				q := obj.Appendp(p, c.newprog)
+				q.As = ASLD
+				q.From.SetConst(int64(shift))
+				q.To = p.To
+				p.From.Offset >>= shift
+				p = q
+			} else if isPPC64DoublewordRotateMask(val) {
+				// This constant is a mask value, generate MOVD $-1, Rto; RLDIC Rto, ^me, mb, Rto
+				mb, me := encodePPC64RLDCMask(val)
+				q := obj.Appendp(p, c.newprog)
+				q.As = ARLDC
+				q.AddRestSourceConst((^int64(me)) & 0x3F)
+				q.AddRestSourceConst(int64(mb))
+				q.From = p.To
+				q.To = p.To
+				p.From.Offset = -1
+				p = q
+			} else {
+				// Load the constant from memory.
+				p.From.Type = obj.TYPE_MEM
+				p.From.Sym = ctxt.Int64Sym(p.From.Offset)
+				p.From.Name = obj.NAME_EXTERN
+				p.From.Offset = 0
+			}
 		}
 	}
 
@@ -107,9 +245,63 @@ func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
 		}
 
 	case ASUB:
-		if p.From.Type == obj.TYPE_CONST {
-			p.From.Offset = -p.From.Offset
-			p.As = AADD
+		if p.From.Type != obj.TYPE_CONST {
+			break
+		}
+		// Rewrite SUB $const,... into ADD $-const,...
+		p.From.Offset = -p.From.Offset
+		p.As = AADD
+		// This is now an ADD opcode, try simplifying it below.
+		fallthrough
+
+	// Rewrite ADD/OR/XOR/ANDCC $const,... forms into ADDIS/ORIS/XORIS/ANDISCC
+	case AADD:
+		// Don't rewrite if this is not adding a constant value, or is not an int32
+		if p.From.Type != obj.TYPE_CONST || p.From.Offset == 0 || int64(int32(p.From.Offset)) != p.From.Offset {
+			break
+		}
+		if p.From.Offset&0xFFFF == 0 {
+			// The constant can be added using ADDIS
+			p.As = AADDIS
+			p.From.Offset >>= 16
+		} else if buildcfg.GOPPC64 >= 10 {
+			// Let the assembler generate paddi for large constants.
+			break
+		} else if (p.From.Offset < -0x8000 && int64(int32(p.From.Offset)) == p.From.Offset) || (p.From.Offset > 0xFFFF && p.From.Offset < 0x7FFF8000) {
+			// For a constant x, 0xFFFF (UINT16_MAX) < x < 0x7FFF8000 or -0x80000000 (INT32_MIN) <= x < -0x8000 (INT16_MIN)
+			// This is not done for 0x7FFF < x < 0x10000; the assembler will generate a slightly faster instruction sequence.
+			//
+			// The constant x can be rewritten as ADDIS + ADD as follows:
+			//     ADDIS $x>>16 + (x>>15)&1, rX, rY
+			//     ADD   $int64(int16(x)), rY, rY
+			// The range is slightly asymmetric as 0x7FFF8000 and above overflow the sign bit, whereas for
+			// negative values, this would happen with constant values between -1 and -32768 which can
+			// assemble into a single addi.
+			is := p.From.Offset>>16 + (p.From.Offset>>15)&1
+			i := int64(int16(p.From.Offset))
+			p.As = AADDIS
+			p.From.Offset = is
+			q := obj.Appendp(p, c.newprog)
+			q.As = AADD
+			q.From.SetConst(i)
+			q.Reg = p.To.Reg
+			q.To = p.To
+			p = q
+		}
+	case AOR:
+		if p.From.Type == obj.TYPE_CONST && uint64(p.From.Offset)&0xFFFFFFFF0000FFFF == 0 && p.From.Offset != 0 {
+			p.As = AORIS
+			p.From.Offset >>= 16
+		}
+	case AXOR:
+		if p.From.Type == obj.TYPE_CONST && uint64(p.From.Offset)&0xFFFFFFFF0000FFFF == 0 && p.From.Offset != 0 {
+			p.As = AXORIS
+			p.From.Offset >>= 16
+		}
+	case AANDCC:
+		if p.From.Type == obj.TYPE_CONST && uint64(p.From.Offset)&0xFFFFFFFF0000FFFF == 0 && p.From.Offset != 0 {
+			p.As = AANDISCC
+			p.From.Offset >>= 16
 		}
 
 	// To maintain backwards compatibility, we accept some 4 argument usage of
@@ -643,7 +835,7 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 
 			q = p
 
-			if NeedTOCpointer(c.ctxt) && c.cursym.Name != "runtime.duffzero" && c.cursym.Name != "runtime.duffcopy" {
+			if NeedTOCpointer(c.ctxt) && !isNOTOCfunc(c.cursym.Name) {
 				// When compiling Go into PIC, without PCrel support, all functions must start
 				// with instructions to load the TOC pointer into r2:
 				//
@@ -675,11 +867,12 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 				q.Pos = p.Pos
 				q.From.Type = obj.TYPE_CONST
 				q.From.Offset = 0x38420000
-				rel := obj.Addrel(c.cursym)
-				rel.Off = 0
-				rel.Siz = 8
-				rel.Sym = c.ctxt.Lookup(".TOC.")
-				rel.Type = objabi.R_ADDRPOWER_PCREL
+				c.cursym.AddRel(c.ctxt, obj.Reloc{
+					Type: objabi.R_ADDRPOWER_PCREL,
+					Off:  0,
+					Siz:  8,
+					Sym:  c.ctxt.Lookup(".TOC."),
+				})
 			}
 
 			if !c.cursym.Func().Text.From.Sym.NoSplit() {
@@ -777,15 +970,15 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 			if c.cursym.Func().Text.From.Sym.Wrapper() {
 				// if(g->panic != nil && g->panic->argp == FP) g->panic->argp = bottom-of-frame
 				//
-				//	MOVD g_panic(g), R3
-				//	CMP R0, R3
+				//	MOVD g_panic(g), R22
+				//	CMP R22, $0
 				//	BEQ end
-				//	MOVD panic_argp(R3), R4
-				//	ADD $(autosize+8), R1, R5
-				//	CMP R4, R5
+				//	MOVD panic_argp(R22), R23
+				//	ADD $(autosize+8), R1, R24
+				//	CMP R23, R24
 				//	BNE end
-				//	ADD $8, R1, R6
-				//	MOVD R6, panic_argp(R3)
+				//	ADD $8, R1, R25
+				//	MOVD R25, panic_argp(R22)
 				// end:
 				//	NOP
 				//
@@ -804,9 +997,9 @@ func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
 				q = obj.Appendp(q, c.newprog)
 				q.As = ACMP
 				q.From.Type = obj.TYPE_REG
-				q.From.Reg = REG_R0
-				q.To.Type = obj.TYPE_REG
-				q.To.Reg = REG_R22
+				q.From.Reg = REG_R22
+				q.To.Type = obj.TYPE_CONST
+				q.To.Offset = 0
 
 				q = obj.Appendp(q, c.newprog)
 				q.As = ABEQ
@@ -1363,11 +1556,12 @@ func (c *ctxt9) stacksplit(p *obj.Prog, framesize int32) *obj.Prog {
 		p.To.Reg = REG_R2
 	}
 
+	// The instructions which unspill regs should be preemptible.
+	p = c.ctxt.EndUnsafePoint(p, c.newprog, -1)
 	unspill := c.cursym.Func().UnspillRegisterArgs(p, c.newprog)
-	p = c.ctxt.EndUnsafePoint(unspill, c.newprog, -1)
 
 	// BR	start
-	p = obj.Appendp(p, c.newprog)
+	p = obj.Appendp(unspill, c.newprog)
 	p.As = ABR
 	p.To.Type = obj.TYPE_BRANCH
 	p.To.SetTarget(startPred.Link)
