@@ -11,9 +11,8 @@ import (
 	"internal/testenv"
 	"os"
 	"os/exec"
-	"reflect"
 	"runtime/metrics"
-	"sort"
+	"slices"
 	"strings"
 	"testing"
 )
@@ -110,12 +109,15 @@ func TestCmdBisect(t *testing.T) {
 
 	var want []string
 	src, err := os.ReadFile("godebug_test.go")
+	if err != nil {
+		t.Fatal(err)
+	}
 	for i, line := range strings.Split(string(src), "\n") {
 		if strings.Contains(line, "BISECT"+" "+"BUG") {
 			want = append(want, fmt.Sprintf("godebug_test.go:%d", i+1))
 		}
 	}
-	sort.Strings(want)
+	slices.Sort(want)
 
 	var have []string
 	for _, line := range strings.Split(string(out), "\n") {
@@ -123,9 +125,9 @@ func TestCmdBisect(t *testing.T) {
 			have = append(have, line[strings.LastIndex(line, "godebug_test.go:"):])
 		}
 	}
-	sort.Strings(have)
+	slices.Sort(have)
 
-	if !reflect.DeepEqual(have, want) {
+	if !slices.Equal(have, want) {
 		t.Errorf("bad bisect output:\nhave %v\nwant %v\ncomplete output:\n%s", have, want, string(out))
 	}
 }
